/*
 *		PORT I/O functions
 *			coded by Kenar
 *
 *		In this file cread && cwrite are defined explicitly.
 *		However they are renamed as 
 *			pread && pwrite
 *		in 'tip.h'.
 */

#include <nje/nje.h>
#include "tip.h"

#define is_kshjis(x) ((lett)(x) >= ((lett)(0xa0)) && (lett)(x) < (lett)(0xe0))
#define rread(buf,n) (*_rread)(buf,n)
#define rwrite(buf,n) (*_rwrite)(buf,n)

char *strlwr(char*);
static char partab[0200];
static int (*_rread)(),(*_rwrite)();



static int raw8_read(buf,bufsiz) /* added by Kenar */
char *buf;
int bufsiz;
{	return read(FD, buf, bufsiz);
}

static int raw8_write(buf,n)
char *buf;
int n;
{	return write(FD,buf,n);
}


static int raw7_read(buf,bufsiz) /* added by Kenar */
char *buf;
int bufsiz;
{	int n,k;
	n = read(FD, buf, bufsiz);
	for(k = 0; k < n; k++) buf[k] &= 0x7F;
	return n;
}

static int raw7_write(buf,n) /* added by Kenar */
char *buf;
int n;
{	int m,k;
	for(k = 0; k < n; k++) buf[k] = partab[buf[k] & 0x7f];
		/* turn parity bit using current parity table */
	m = write(FD, buf, n);
		
	for(k = 0; k < n; k++) buf[k] &= 0x7f;	/* resume buf */
	return m;
}


static int movechar(buf,xbuf,k,n)
lett *buf,*xbuf;
int k,n;
/*	Movechar moves first n chars in xbuf to buf,
	and them shift the rest in xbuf to the left size.
	Here k is the effective size of xbuf.
	Movechar returns actually meved char count.
*/
{	int i;
	if(k < n) n = k;
	for(i = 0; i < n; i++) buf[i] = xbuf[i];
	for(i = n; i < k; i++) xbuf[i - n] = xbuf[i]; // pack to left
	return n;
}

lett *memapd(dst,src,len)
lett *dst,*src;
int len;
{	while(len--) *dst++ = *src++;
	return dst;
}


/* JIS escape sequence */
static const lett *KI = "\033$B";   // JIS83
//static const lett *KI = "\033$@"; // OLD JIS
static const lett *KO = "\033(J";   // JIS83
//static const lett *KO = "\033(H"; // OLD JIS
//static const lett *KO="\33(B";    // JIS used in jconv
static const lett SO = 0x0e;
static const lett SI = 0x0f;
#define sizeKI strlen(KI)
#define sizeKO strlen(KO)

static int _jis_read(buf,bufsiz) /* added by Kenar */
lett *buf;
int bufsiz;
/*	Jis_read reads JIS code from port FD at most bufsiz chars,
	and translates them to euc code and then moves them to buf.
	_jis_read may return 0.
						- coded by Kenar -
*/
{	int n;
	static lett xbuf[BUFSIZ];
	static int kana = NO;
	static int kanji = NO;
	static int i = 0;
	static int k = 0;
	lett *s,*p;
	
	if(k == 0)
	{	n = rread(buf, bufsiz);
		if(n < 1) return -1;
	}
	else 
	{	k -= n = movechar(buf,xbuf,k,bufsiz);
		return n;
	}

	s = buf;
	p = xbuf;
	while(n-- > 0)
	{	if(*s == SO) /* to kana mode */
			kana = YES;
		else if(*s == SI) /* to ascii mode */
			kana = NO;
		else if(kana && !iscntrl(*s) && *s != ' ')
		{	*p++ = SS2;
			*p++ = *s | 0x80;
		}
		else if(kanji)
		{	if(*s == KO[i])  /* being out from kanji mode */
			{	i++;
				if(i == sizeKO)
				{	i = 0;
					kanji = NO;
				}
			}
			else if(i > 0) /* flush sbuf */
			{	p = memapd(p,KO,i);
				i = 0;
			}
			else if(*s <= ' ') *p++ = *s;
			else *p++ = *s | 0x80;
		}
		else
		{	if(*s == KI[i]) /* being into kanji mode */
			{	i++;
				if(i == sizeKI)
				{	i = 0;
					kanji = YES;
				}
			}
			else if(i > 0) /* flush sbuf */
			{	p = memapd(p,KI,i);
				i = 0;
			}
			else *p++ = *s;
		}
		s++;
	}
	
	k = p - xbuf;
	n = movechar(buf,xbuf,k,bufsiz);
		
	k -= n;
	/*	n is the chars to be moved to buf.
		The count of chars in xbuf is then k.
	   	xbuf is already converted to EUC.
		If(k<n) then only k bytes are moved.
	*/
	return n;
}

static int jis_read(buf,bufsiz) /* added by Kenar */
lett *buf;
int bufsiz;
/*	Jis_read reads JIS code from port FD at most bufsiz chars,
	and translates them to euc code and then moves them to buf.
	jis_read read at least one char.
						- coded by Kenar -
*/
{	int n;
	repeat
	{	n = _jis_read(buf,bufsiz);
	} until( n > 0 );
	return n;
}



static int jis_write(buf,n) /* added by Kenar */
lett *buf;
int n; // effective count of chars in buf
/*	Jis_write reads EUC codes in buf and translates them to JIS.
	And then writes it to port FD.
						- coded by Kenar -
*/
{	lett *p,*s,xbuf[BUFSIZ];
	static lett sbuf[3]; /* sbuf has a rest of data */
	static int m = 0; // Char count left unresolved
	static int i = 0; // Code count in sbuf
	static int kana = NO;
	static int kanji = NO;
	p = buf;
	s = xbuf;
	/* The EUC chars in buf may be a incomplete EUC code,
	   that is, it may contain a part of EUC at the end.
	   They must be saved sbuf until next operation.
	   We store the JIS code temporary to xbuf.
	*/
	while(n > 0)
	{	if(m == 0) /* then we are out of extended code */
		{	if(*p & 0x80) /* beginning of extended code */
			{	if(*p == SS2) /* code set 2 (KANA) */
				{	sbuf[0] = SS2;
					m = 1;
				}
				else if(*p == SS3) /* code set 3 */
				{	sbuf[0] = SS3;
					m = 2;
				}
				else /* code set 1 */
				{	sbuf[0] = *p;
					m = 1;
				}
				i++;
			}
			else /* code set 0 */
			{	if(kana)
				{	kana = NO;
					*s++ = SI;
				}
				if(kanji)
				{	kanji = NO;
					s = memapd(s,KO,sizeKO);
				}
				*s++ = *p;
			}
		}
		else	/* in extended code */
		{	sbuf[i++] = *p;
			m--;
			if(m == 0)/* then we have got complete extended code */
			{	if(sbuf[0] == SS2) /* KANA code */
				{	if(kanji)
					{	s = memapd(s,KO,sizeKO);
						kanji = NO;
					}
					if(!kana)
					{	kana = YES;
						*s++ = SO;
					}
					*s++ = sbuf[1] & 0x7f;
				}
				else if(sbuf[0] == SS3); /* ignore */ 
				else /* KANJI code */
				{	if(kana)
					{	kana = NO;
						*s++ = SI;
					}
					if(!kanji)
					{	kanji = YES;
						s = memapd(s,KI,sizeKI);
					}
					*s++ = sbuf[0] & 0x7f;
					*s++ = sbuf[1] & 0x7f;
				}
				i = 0;
			}
		}
				
		p++;
		n--;
	}
	/* Then chars in xbuf are complete SJIS code */
	if(s > xbuf) rwrite(xbuf, s - xbuf);
}





static void sjis_euc(c1,c2,ptr)
lett c1,c2,*ptr;
/* Quated from "alex113f" by Mr.Kasuga */
{
        if (c1 >= 0xe0)
                c1 -= 0x40;
        if (c2 >= 0x9f) {
                *ptr++ = (c1 - 0x88)*2 + 0xb0;
                *ptr = c2 + 0x02;
        }
        else {
                if (c2 >= 0x7f)
                        c2 -= 1;
                *ptr++ = (c1 - 0x89)*2 + 0xb1;
                *ptr = c2 + 0x61;
        }
}



static int sjis_read(buf,bufsiz) /* added by Kenar */
lett *buf;
int bufsiz;
/*	Sjis_read reads sjis code from port FD at most bufsiz chars,
	and translates them to euc code and then moves them to buf.
	If bufsiz == 1 then sjis_read reads one char.
						- coded by Kenar -
*/
{	int n,i;
	static k = 0;
	lett ub,lb;
		// ub: upper byte of shift-jis and a flag in shift-jis
		// lb: lower byte of shift-jis
	static lett xbuf[BUFSIZ];
		/* The old code including bug are as followos:	*/
		/*	if(k < bufsiz)				*/
		/*		n = read(FD, buf, bufsiz - k);	*/
		/*	else					*/
		/*	{	k -= movechar(buf,xbuf,k,bufsiz);*/
		/*		return bufsiz;			*/
		/*	}					*/
		/* The data in buf will be appended to 'xbuf'.  */
		/* But this will prevent flush buffer if 'buf'  */
		/* are not filled  from port. 			*/

	if(k == 0)
	{	n = read(FD, buf, bufsiz);
		if(n <= 0) return n;
	}
	else 
	{	k -= n = movechar(buf,xbuf,k,bufsiz);
		return n;
	}
	ub = 0;

	i = 0;
 	// i controls the reading point of buf
	// on the other hand, k controls the writing point of xbuf
	while(i < n)
	{	if(!ub && is_ushjis(buf[i])) ub = buf[i];
		else if(ub) // This is the case that buf[i] is the lower byte
		{	sjis_euc(ub,buf[i],&xbuf[k]);
			k += 2;
			ub = 0;
		}
		else if(is_kshjis(buf[i])) // KANA of shift jis
		{	xbuf[k] = SS2;
                        xbuf[k+1] = buf[i];
			k += 2;
		}
		else
		{	xbuf[k] = buf[i];
			k++;
		}
		i++;
	}
	/* At this stage the size of xbuf is k */
	if(ub)
	{	read(FD,&lb,1);
		sjis_euc(ub,lb,&xbuf[k]);
		k += 2;
	}
	n = movechar(buf,xbuf,k,bufsiz);
	k -= n;
	/*	n is the chars to be moved to buf.
		The count of chars in xbuf is then k.
	   	xbuf is already converted to EUC.
		If(k<n) then only k bytes are moved.
	*/
	return n;
}

static void euc_sjis(c1,c2,ptr)
/* Quated from "alex113f" by Mr.Kasuga */
lett c1,c2,*ptr;
{	c1 &= 0x7f;
	c2 &= 0x7f;
	
	
	if (c1 >= 0x5f)
		c1 += 0x80;
	if ((c1 % 2) == 0) {
		*ptr++ = ((c1 - 0x30)/2 + 0x88) & 0xff;
		*ptr = (c2 + 0x7e) & 0xff;
	}
	else {
		if (c2 >= 0x60)
			c2 += 1;
		*ptr++ = ((c1 - 0x31)/2 + 0x89) & 0xff;
		*ptr = (c2 + 0x1f) & 0xff;
	}
}

static int sjis_write(buf,n) /* added by Kenar */
lett *buf;
int n; // effective count of chars in buf
/*	Sjis_write reads EUC codes in buf and translates them to SJIS.
	And then writes it to port FD.
						- coded by Kenar -
*/
{	lett *p,*s,xbuf[BUFSIZ];
	static lett sbuf[3]; /* sbuf has a rest of data */
	static int m = 0; // Char count left unresolved
	static int i = 0; // Code count in sbuf
	p = buf;
	s = xbuf;
	/* The EUC chars in buf may be a incomplete EUC code,
	   that is, it may contain a part of EUC at the end.
	   They must be saved sbuf until next operation.
	   We store the SJIS code temporary to xbuf.
	*/
	while(n > 0)
	{	if(m == 0) /* then we are out of extended code */
		{	if(*p & 0x80) /* beginning of extended code */
			{	if(*p == SS2) /* code set 2 (KANA) */
				{	sbuf[0] = SS2;
					m = 1;
				}
				else if(*p == SS3) /* code set 3 */
				{	sbuf[0] = SS3;
					m = 2;
				}
				else /* code set 1 */
				{	sbuf[0] = *p;
					m = 1;
				}
				i++;
			}
			else /* code set 0 */
				*s++ = *p;
		}
		else	/* in extended code */
		{	sbuf[i++] = *p;
			m--;
			if(m == 0)/* then we have got complete extended code */
			{	if(sbuf[0] == SS2)
					*s++ = sbuf[1]; /* KANA code */
				else if(sbuf[0] == SS3); /* ignore */ 
				else /* SJIS */
				{	euc_sjis(sbuf[0],sbuf[1],s);
					s += 2;
				}
				i = 0;
			}
		}
		p++;
		n--;
	}
	/* Then chars in xbuf are complete SJIS code */
	if(s > xbuf) write(FD, xbuf, s - xbuf);
}



/*
 * setparity: Build a parity table with appropriate high-order bit.
 *
 *	COMMENT by Kenar:
 * 	The main part of this routine is transfered from tip.c
 * 	Parity table defined in partab.c. was discarded,
 *	and in stead, constructed here.
 */
 
void setparity()
{
	register int i;
	char *parity;
	
	if(value(PARITY) == NOSTR)
		value(PARITY) = "none";	// default parity
	parity = value(PARITY);
	
	/* added ky Kenar */
	if(equal(parity,"none"))
	{	_rread = raw8_read;
		_rwrite = raw8_write;
		return;
	}
		
	/* added by Kenar */
	/* Construct parity table */
	/* At first time we assume even parity */
	{	int x,count;
		for ( i = 0; i < 0200; i++)
		{	x = i;
			count = 0;
			while(x)
			{	if(x & 1) count++;
				x >>= 1;
			}
			if(count & 1) partab[i] = (i | 0200) ;
			else partab[i] = i;
		}
	}
	
	
	if(equal(parity,"odd"))
	{
		for(i = 0; i < 0200; i++)
			partab[i] ^= 0200;	/* reverse bit 7 */
	}
	else if(equal(parity,"zero"))
	{
		for(i = 0; i < 0200; i++)
			partab[i] &= ~0200;	/* turn off bit 7 */
	}
	else if(equal(parity,"one"))
	{
		for(i = 0; i < 0200; i++)
			partab[i] |= 0200;	/* turn on bit 7 */
	}
	else if(!equal(parity,"even"))
	{	fprintf(stderr,
			"%s: Unknown parity. Even is assumed.\n",parity);
		fflush(stderr);
		value(PARITY) = makstr("even");
	}
	_rread = raw7_read;
	_rwrite = raw7_write;
}

void setpio() /* added by Kenar */
{	char lt[20];
	if(equal(value(LANGTYPE),"")) value(LANGTYPE) = "ascii";
	strcpy(lt,value(LANGTYPE));
	strlwr(lt);
	if(equal(lt,"euc"))
	{	cread = raw8_read;
		cwrite = raw8_write;
	}
	else if(equal(lt,"jis"))
	{	cread = jis_read;
		cwrite = jis_write;
	}
	else if(equal(lt,"sjis"))
	{	cread = sjis_read;
		cwrite = sjis_write;
	}
	else
	{	cread = _rread;
		cwrite = _rwrite;
		if(!equal(lt,"ascii"))
		{	printf
			("%s: Undefined language type. ASCII is assumed\n",lt);
			value(LANGTYPE) = makstr("ascii");
		}
			
	}
}
